@@ -77,8 +77,8 @@ use vars	qw{
 $TIMELOG->{dummy}=0;
 @ISA = qw();
 
-$VERSION = "2.30";
-$SUB_VERSION = 'devel $Id: SpamAssassin.pm,v 1.94 2002/06/14 23:17:15 hughescr Exp $';
+$VERSION = "2.31";
+$SUB_VERSION = 'devel $Id: SpamAssassin.pm,v 1.94.2.2 2002/06/20 17:20:29 hughescr Exp $';
 
 sub Version { $VERSION; }
 
@@ -725,11 +725,23 @@ sub create_default_prefs {
 
 sub expand_name ($) {
   my ($self, $name) = @_;
-  if ($name eq '') {
-      return $ENV{HOME} if ($ENV{HOME} && $ENV{HOME} =~ /\//);
-      return (getpwuid($>))[7];
-  } else {
-      return (getpwnam($name))[7];
+  my $home = $ENV{'HOME'} || '';
+
+  if ($^O =~ /mswin|dos|os2/oi) {
+	  my $userprofile = $ENV{'USERPROFILE'} || '';
+
+	  return $userprofile if ($userprofile && $userprofile =~ m/^[a-z]\:[\/\\]/oi);
+	  return $userprofile if ($userprofile =~ m/^\\\\/o);
+
+	  return $home if ($home && $home =~ m/^[a-z]\:[\/\\]/oi);
+	  return $home if ($home =~ m/^\\\\/o);
+
+	  return '';
+  }
+  else {
+	  return $home if ($home && $home =~ /\//o);
+	  return (getpwnam($name))[7] if ($name ne '');
+	  return (getpwuid($>))[7];
   }
 }
 
