@@ -168,9 +168,13 @@ sub get_header {
   if (!wantarray) {
       # If there is no header like that, return undef
       if (scalar(@entries) < 1 ) { return undef; }
-      foreach my $entry (@entries)
-      {
-	  if($entry->{count} > 0) { return $entry->{0}; }
+      foreach my $entry (@entries) {
+	  if($entry->{count} > 0) {
+	    my $ret = $entry->{0};
+            $ret =~ s/^\s+//;
+            $ret =~ s/\n\s+/ /g;
+	    return $ret;
+	  }
       }
       return undef;
 
@@ -182,7 +186,12 @@ sub get_header {
       # loop through each entry and collect all the individual matching lines
       foreach my $entry (@entries)
       {
-	  foreach my $i (0 .. ($entry->{count}-1)) { push (@ret, $entry->{$i}); }
+	  foreach my $i (0 .. ($entry->{count}-1)) {
+		my $ret = $entry->{$i};
+                $ret =~ s/^\s+//;
+                $ret =~ s/\n\s+/ /g;
+	  	push (@ret, $ret);
+          }
       }
 
       return @ret;
@@ -206,14 +215,108 @@ sub get_all_headers {
   if (!defined ($self->{add_From_line}) || $self->{add_From_line} == 1) {
     my $from = $self->{from_line};
     if (!defined $from) {
-      my $f = $self->get_header("From"); $f ||= "spamassassin\@localhost\n";
+      my $f = $self->get_header("From") || '';
       chomp ($f);
 
-      $f =~ s/^.*?<(.+)>\s*$/$1/g               # Foo Blah <jm@foo>
-          or $f =~ s/^(.+)\s\(.*?\)\s*$/$1/g;   # jm@foo (Foo Blah)
+      #warn "$f";
+      #warn "---";
+
+      $f =~ tr/\n/ /s;                            # unwrap the line
+      $f =~ s/^\s*(.*?)\s*$/$1/;                  # and remove leading and trailing spaces
+
+      #warn "$f";
+      #warn "---";
+
+      if ($f) {
+        my $p = -1;                               # position of block opening
+        my $s = '';                               # closing block seperator; either a quote or a bracket
+        my $n = 0;                                # deepness of nesting (only for brackets)
+        for (my $i = 0; $i < length($f); $i++) {
+          my $c = substr($f, $i, 1);
+          #warn "$f, $i, $c, $p, $s, $n";
+          if ($p == -1) {
+            if ($c eq '"' or $c eq '(') {         # find next block opening
+              $s = $c eq '(' ? ')' : '"';
+              $p = $i;
+            }
+            elsif ($c eq ',') {                   # begin of next address;
+              substr($f, $i) = '';                # look at the first one only
+              last;                               # and break here
+            }
+          }
+          else {
+            if ($c eq '(') {                      # a nested block is opened
+              $n++;
+            }
+            elsif ($c eq '\\') {                  # the next character is escaped,
+              $i++;                               # so skip it
+              next;
+            }
+            elsif ($c eq $s) {                    # block closing;
+              if ($n and $s eq ')') {             # it's a nested block
+                $n--;
+                next;
+              }
+              else {                              # it's no nested block,
+                substr($f, $p, $i - $p + 1) = ''; # so remove it
+                $i = $p - 1;                      # and proceed where it started
+                $p = -1;
+                next;
+              }
+            }
+          }
+        }
+        if ($p > -1) {                            # block not closed,
+          substr($f, $p) = '';                    # so remove till the end
+        }
+      }
+      $f =~ s/^\s*(.*?)\s*$/$1/;                  # remove leading and trailing spaces again
+
+      #warn "$f";
+      #warn "---";
+
+      if ($f) {
+        my $p = index($f, '<');
+        if ($p > -1) {                            # seems like we've got angle brackets
+          substr($f, 0, $p + 1) = '';             # remove everything in front
+          $p = index($f, '>');
+          if ($p > -1) {
+            $f = substr($f, 0, $p);               # get the contents
+            if ($f =~ /\s/) {                     # address contains spaces;
+              $f = '';                            # illegal address
+            }
+          }
+          else {                                  # no closing angle bracket;
+            $f = '';                              # illegal address
+          }
+        }
+        else {                                    # no angle bracket,
+          $f =~ s/\s.*//;                         # so remove everything except the first token
+        }
+      }
+
+      #warn "$f";
+      #warn "---";
+
+      if ($f) {
+        my $p = index($f, '@');
+        if (   $p == 0                            # we either lost the local part,
+            or $p == length($f) - 1               # or the domain
+            or $p != rindex($f, '@')) {           # or it contains more than one @;
+          $f = '';                                # illegal address anyway
+        }
+      }
+
+      unless ($f) {                               # we lost the address;
+        $f = 'spamassassin@localhost';            # use a default one
+      }
+
+      #warn "$f";
+      #warn "---";
+
       $from = "From $f  ".(scalar localtime(time))."\n";
     }
-    push (@lines, $from);
+    unshift (@lines, $from);
   }
 
   foreach my $hdrcode (@{$self->{header_order}}) {
