@@ -35,11 +35,17 @@
 #endif
 
 /* SunOS 4.1.4 patch from Tom Lipkis <tal@pss.com> */
-#if (defined(__sun__) && defined(__sparc__) && !defined(__svr4__)) /* SunOS */
-     || (defined(_IRIX_))  /* IRIX */
-     || (defined(__osf__)) /* Digital UNIX */
+#if (defined(__sun__) && defined(__sparc__) && !defined(__svr4__)) /* SunOS */ \
+     || (defined(__sgi))  /* IRIX */ \
+     || (defined(__osf__)) /* Digital UNIX */ \
+     || (defined(hpux) || defined(__hpux)) /* HPUX */
+# ifndef h_errno
+# define h_errno errno
+# endif
+
 # ifndef EX__MAX
 # define EX__MAX 77
+
 extern char *optarg;
 typedef unsigned long	in_addr_t;	/* base type for internet address */
 # endif
